@@ -1,7 +1,5 @@
 # BSD 3-Clause License; see https://github.com/scikit-hep/awkward-1.0/blob/main/LICENSE
 
-from __future__ import absolute_import
-
 import json
 import copy
 import math
@@ -30,56 +28,62 @@ class BitMaskedArray(Content):
         lsb_order,
         identifier=None,
         parameters=None,
+        nplike=None,
     ):
         if not (isinstance(mask, Index) and mask.dtype == np.dtype(np.uint8)):
             raise TypeError(
-                "{0} 'mask' must be an Index with dtype=uint8, not {1}".format(
+                "{} 'mask' must be an Index with dtype=uint8, not {}".format(
                     type(self).__name__, repr(mask)
                 )
             )
         if not isinstance(content, Content):
             raise TypeError(
-                "{0} 'content' must be a Content subtype, not {1}".format(
+                "{} 'content' must be a Content subtype, not {}".format(
                     type(self).__name__, repr(content)
                 )
             )
         if not isinstance(valid_when, bool):
             raise TypeError(
-                "{0} 'valid_when' must be boolean, not {1}".format(
+                "{} 'valid_when' must be boolean, not {}".format(
                     type(self).__name__, repr(valid_when)
                 )
             )
-        if not ak._util.isint(length):
-            raise TypeError(
-                "{0} 'length' must be an integer, not {1}".format(
-                    type(self).__name__, repr(length)
+        if not isinstance(length, ak._v2._typetracer.UnknownLengthType):
+            if not (ak._util.isint(length) and length >= 0):
+                raise TypeError(
+                    "{} 'length' must be a non-negative integer, not {}".format(
+                        type(self).__name__, length
+                    )
                 )
-            )
         if not isinstance(lsb_order, bool):
             raise TypeError(
-                "{0} 'lsb_order' must be boolean, not {1}".format(
+                "{} 'lsb_order' must be boolean, not {}".format(
                     type(self).__name__, repr(lsb_order)
                 )
             )
-        if not length <= len(mask) * 8:
+        if length > mask.length * 8:
             raise ValueError(
-                "{0} 'length' ({1}) must be <= len(mask) * 8 ({2})".format(
-                    type(self).__name__, length, len(mask) * 8
+                "{} 'length' ({}) must be <= len(mask) * 8 ({})".format(
+                    type(self).__name__, length, mask.length * 8
                 )
             )
-        if not length <= len(content):
+        if length > content.length:
             raise ValueError(
-                "{0} 'length' ({1}) must be <= len(content) ({2})".format(
-                    type(self).__name__, length, len(content)
+                "{} 'length' ({}) must be <= len(content) ({})".format(
+                    type(self).__name__, length, content.length
                 )
             )
+        if nplike is None:
+            nplike = content.nplike
+        if nplike is None:
+            nplike = mask.nplike
 
         self._mask = mask
         self._content = content
         self._valid_when = valid_when
         self._length = length
         self._lsb_order = lsb_order
-        self._init(identifier, parameters)
+        self._init(identifier, parameters, nplike)
 
     @property
     def mask(self):
@@ -97,10 +101,6 @@ class BitMaskedArray(Content):
     def lsb_order(self):
         return self._lsb_order
 
-    @property
-    def nplike(self):
-        return self._mask.nplike
-
     Form = BitMaskedForm
 
     def _form_with_key(self, getkey):
@@ -118,24 +118,39 @@ class BitMaskedArray(Content):
     def _to_buffers(self, form, getkey, container, nplike):
         assert isinstance(form, self.Form)
         key = getkey(self, form, "mask")
-        container[key] = ak._v2._util.little_endian(self._mask.to(nplike))
+        container[key] = ak._v2._util.little_endian(self._mask.raw(nplike))
         self._content._to_buffers(form.content, getkey, container, nplike)
 
     @property
     def typetracer(self):
+        tt = ak._v2._typetracer.TypeTracer.instance()
         return BitMaskedArray(
-            ak._v2.index.Index(self._mask.to(ak._v2._typetracer.TypeTracer.instance())),
+            ak._v2.index.Index(self._mask.raw(tt)),
             self._content.typetracer,
             self._valid_when,
             self._length,
             self._lsb_order,
             self._typetracer_identifier(),
             self._parameters,
+            tt,
         )
 
-    def __len__(self):
+    @property
+    def length(self):
         return self._length
 
+    def _forget_length(self):
+        return BitMaskedArray(
+            self._mask,
+            self._content.typetracer,
+            self._valid_when,
+            ak._v2._typetracer.UnknownLength,
+            self._lsb_order,
+            self._identifier,
+            self._parameters,
+            self._nplike,
+        )
+
     def __repr__(self):
         return self._repr("", "", "")
 
@@ -145,7 +160,7 @@ class BitMaskedArray(Content):
         out.append(" lsb_order=")
         out.append(repr(json.dumps(self._lsb_order)))
         out.append(" len=")
-        out.append(repr(str(len(self))))
+        out.append(repr(str(self.length)))
         out.append(">")
         out.extend(self._repr_extra(indent + "    "))
         out.append("\n")
@@ -164,20 +179,21 @@ class BitMaskedArray(Content):
             self._lsb_order,
             self._identifier,
             ak._v2._util.merge_parameters(self._parameters, parameters),
+            self._nplike,
         )
 
     def toByteMaskedArray(self):
-        nplike = self._mask.nplike
-        bytemask = ak._v2.index.Index8.empty(len(self._mask) * 8, nplike)
+        bytemask = ak._v2.index.Index8.empty(self._mask.length * 8, self._nplike)
+        assert bytemask.nplike is self._nplike and self._mask.nplike is self._nplike
         self._handle_error(
-            nplike[
+            self._nplike[
                 "awkward_BitMaskedArray_to_ByteMaskedArray",
                 bytemask.dtype.type,
                 self._mask.dtype.type,
             ](
-                bytemask.to(nplike),
-                self._mask.to(nplike),
-                len(self._mask),
+                bytemask.data,
+                self._mask.data,
+                self._mask.length,
                 False,  # this differs from the kernel call in 'bytemask'
                 self._lsb_order,
             )
@@ -188,20 +204,21 @@ class BitMaskedArray(Content):
             self._valid_when,
             self._identifier,
             self._parameters,
+            self._nplike,
         )
 
     def toIndexedOptionArray64(self):
-        nplike = self._mask.nplike
-        index = ak._v2.index.Index64.empty(len(self._mask) * 8, nplike)
+        index = ak._v2.index.Index64.empty(self._mask.length * 8, self._nplike)
+        assert index.nplike is self._nplike and self._mask.nplike is self._nplike
         self._handle_error(
-            nplike[
+            self._nplike[
                 "awkward_BitMaskedArray_to_IndexedOptionArray",
                 index.dtype.type,
                 self._mask.dtype.type,
             ](
-                index.to(nplike),
-                self._mask.to(nplike),
-                len(self._mask),
+                index.raw(self._nplike),
+                self._mask.data,
+                self._mask.length,
                 self._valid_when,
                 self._lsb_order,
             ),
@@ -211,24 +228,26 @@ class BitMaskedArray(Content):
             self._content,
             self._identifier,
             self._parameters,
+            self._nplike,
         )
 
     def mask_as_bool(self, valid_when=None, nplike=None):
         if valid_when is None:
             valid_when = self._valid_when
         if nplike is None:
-            nplike = self._mask.nplike
+            nplike = self._nplike
 
-        bytemask = ak._v2.index.Index8.empty(len(self._mask) * 8, nplike)
+        bytemask = ak._v2.index.Index8.empty(self._mask.length * 8, nplike)
+        assert bytemask.nplike is self._nplike and self._mask.nplike is self._nplike
         self._handle_error(
             nplike[
                 "awkward_BitMaskedArray_to_ByteMaskedArray",
                 bytemask.dtype.type,
                 self._mask.dtype.type,
             ](
-                bytemask.to(nplike),
-                self._mask.to(nplike),
-                len(self._mask),
+                bytemask.data,
+                self._mask.data,
+                self._mask.length,
                 0 if valid_when == self._valid_when else 1,
                 self._lsb_order,
             )
@@ -239,9 +258,12 @@ class BitMaskedArray(Content):
         return self._content._getitem_range(slice(0, 0))
 
     def _getitem_at(self, where):
+        if not self._nplike.known_data:
+            return ak._v2._typetracer.MaybeNone(self._content._getitem_at(where))
+
         if where < 0:
-            where += len(self)
-        if not (0 <= where < len(self)) and self.nplike.known_shape:
+            where += self.length
+        if not (0 <= where < self.length) and self._nplike.known_shape:
             raise NestedIndexError(self, where)
         if self._lsb_order:
             bit = bool(self._mask[where // 8] & (1 << (where % 8)))
@@ -264,7 +286,8 @@ class BitMaskedArray(Content):
             self._lsb_order,
             self._field_identifier(where),
             None,
-        )
+            self._nplike,
+        ).simplify_optiontype()
 
     def _getitem_fields(self, where, only_fields=()):
         return BitMaskedArray(
@@ -275,7 +298,8 @@ class BitMaskedArray(Content):
             self._lsb_order,
             self._fields_identifier(where),
             None,
-        )
+            self._nplike,
+        ).simplify_optiontype()
 
     def _carry(self, carry, allow_lazy, exception):
         assert isinstance(carry, ak._v2.index.Index)
@@ -324,16 +348,16 @@ class BitMaskedArray(Content):
                 ak._v2.contents.indexedarray.IndexedArray,
                 ak._v2.contents.indexedoptionarray.IndexedOptionArray,
                 ak._v2.contents.bytemaskedarray.ByteMaskedArray,
-                ak._v2.contents.bitmaskeddarray.BitMaskedArray,
-                ak._v2.contents.unmaskeddarray.UnmaskedArray,
+                ak._v2.contents.bitmaskedarray.BitMaskedArray,
+                ak._v2.contents.unmaskedarray.UnmaskedArray,
             ),
         ):
-            return self.toIndexedOptionArray64.simplify_optiontype
+            return self.toIndexedOptionArray64().simplify_optiontype()
         else:
             return self
 
     def num(self, axis, depth=0):
-        return self.toByteMaskedArray.num(axis, depth)
+        return self.toByteMaskedArray().num(axis, depth)
 
     def _offsets_and_flattened(self, axis, depth):
         return self.toByteMaskedArray._offsets_and_flattened(axis, depth)
@@ -345,7 +369,7 @@ class BitMaskedArray(Content):
         if isinstance(
             other,
             (
-                ak._v2.contents.emptyArray.EmptyArray,
+                ak._v2.contents.emptyarray.EmptyArray,
                 ak._v2.contents.unionarray.UnionArray,
             ),
         ):
@@ -361,7 +385,7 @@ class BitMaskedArray(Content):
                 ak._v2.contents.unmaskedarray.UnmaskedArray,
             ),
         ):
-            self._content.mergeable(other.content, mergebool)
+            return self._content.mergeable(other.content, mergebool)
 
         else:
             return self._content.mergeable(other, mergebool)
@@ -384,14 +408,14 @@ class BitMaskedArray(Content):
         return self.toByteMaskedArray().numbers_to_type(name)
 
     def _is_unique(self, negaxis, starts, parents, outlength):
-        if len(self._mask) == 0:
+        if self._mask.length == 0:
             return True
         return self.toIndexedOptionArray64()._is_unique(
             negaxis, starts, parents, outlength
         )
 
     def _unique(self, negaxis, starts, parents, outlength):
-        if len(self._mask) == 0:
+        if self._mask.length == 0:
             return self
         out = self.toIndexedOptionArray64()._unique(negaxis, starts, parents, outlength)
         if negaxis is None:
@@ -411,8 +435,10 @@ class BitMaskedArray(Content):
         kind,
         order,
     ):
-        if len(self._mask) == 0:
-            return ak._v2.contents.NumpyArray(self.nplike.empty(0, np.int64))
+        if self._mask.length == 0:
+            return ak._v2.contents.NumpyArray(
+                self._nplike.empty(0, np.int64), None, None, self._nplike
+            )
 
         return self.toIndexedOptionArray64()._argsort_next(
             negaxis,
@@ -468,10 +494,10 @@ class BitMaskedArray(Content):
         )
 
     def _validityerror(self, path):
-        if len(self.mask) * 8 < len(self):
-            return 'at {0} ("{1}"): len(mask) * 8 < length'.format(path, type(self))
-        elif len(self._content) < len(self):
-            return 'at {0} ("{1}"): len(content) < length'.format(path, type(self))
+        if self.mask.length * 8 < self.length:
+            return f'at {path} ("{type(self)}"): len(mask) * 8 < length'
+        elif self._content.length < self.length:
+            return f'at {path} ("{type(self)}"): len(content) < length'
         elif isinstance(
             self._content,
             (
@@ -526,6 +552,7 @@ class BitMaskedArray(Content):
                     self._lsb_order,
                     self._identifier,
                     self._parameters if options["keep_parameters"] else None,
+                    self._nplike,
                 )
 
         else:
@@ -560,7 +587,7 @@ class BitMaskedArray(Content):
             next = self.toIndexedOptionArray64()
 
             content = next._content.packed()
-            if len(content) > self._length:
+            if content.length > self._length:
                 content = content[: self._length]
 
             return ak._v2.contents.indexedoptionarray.IndexedOptionArray(
@@ -568,17 +595,18 @@ class BitMaskedArray(Content):
                 content,
                 next._identifier,
                 next._parameters,
+                self._nplike,
             )
 
         else:
             excess_length = int(math.ceil(self._length / 8.0))
-            if len(self._mask) == excess_length:
+            if self._mask.length == excess_length:
                 mask = self._mask
             else:
                 mask = self._mask[:excess_length]
 
             content = self._content.packed()
-            if len(content) > self._length:
+            if content.length > self._length:
                 content = content[: self._length]
 
             return BitMaskedArray(
@@ -589,6 +617,7 @@ class BitMaskedArray(Content):
                 self._lsb_order,
                 self._identifier,
                 self._parameters,
+                self._nplike,
             )
 
     def _to_list(self, behavior):
@@ -596,10 +625,50 @@ class BitMaskedArray(Content):
         if out is not None:
             return out
 
-        mask = self.mask_as_bool(valid_when=True, nplike=numpy)[: self._length]
+        mask = self.mask_as_bool(valid_when=True, nplike=self.nplike)[: self._length]
         content = self._content._to_list(behavior)
         out = [None] * self._length
         for i, isvalid in enumerate(mask):
             if isvalid:
                 out[i] = content[i]
         return out
+
+    def _to_nplike(self, nplike):
+        content = self._content._to_nplike(nplike)
+        mask = self._mask._to_nplike(nplike)
+        return BitMaskedArray(
+            mask,
+            content,
+            valid_when=self._valid_when,
+            length=len(self),
+            lsb_order=self._lsb_order,
+            identifier=self._identifier,
+            parameters=self._parameters,
+            nplike=nplike,
+        )
+
+    def _to_json(
+        self,
+        nan_string,
+        infinity_string,
+        minus_infinity_string,
+        complex_real_string,
+        complex_imag_string,
+    ):
+        out = self._to_json_custom()
+        if out is not None:
+            return out
+
+        mask = self.mask_as_bool(valid_when=True, nplike=self.nplike)[: self._length]
+        content = self._content._to_json(
+            nan_string,
+            infinity_string,
+            minus_infinity_string,
+            complex_real_string,
+            complex_imag_string,
+        )
+        out = [None] * self._length
+        for i, isvalid in enumerate(mask):
+            if isvalid:
+                out[i] = content[i]
+        return out
